package cz.muni.fi.pb162.school;

import java.io.BufferedReader;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.io.Writer;
import java.util.Collection;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import java.util.TreeMap;
import junit.framework.Test;
import junit.framework.TestCase;
import junit.framework.TestSuite;

/**
 * Testovaci trida pro katalog studentu
 *
 * @author Petr Adamek &lt;xadamek2@fi.muni.cz&gt;
 */
public class StudentCatalogTest extends TestCase {
    
    public StudentCatalogTest(String testName) {
        super(testName);
    }

    public static Test suite() {
        TestSuite suite = new TestSuite(StudentCatalogTest.class);
        return suite;
    }

    public static final String ENCODING = "UTF-8";
    
    // Tato metoda vytvori testovaci InputStream, ktery bude obsahovat data 
    // predana jako parametr v kodovani UTF-8
    private static byte[] stringLinesToByteArray(String[] lines) throws IOException {
        ByteArrayOutputStream out = new ByteArrayOutputStream();
        Writer writer = new OutputStreamWriter(out,ENCODING);
        for (String line : lines) {
            writer.write(line);
            writer.write("\n");
        }
        writer.flush();
        return out.toByteArray();
    }

    private static InputStream createTestInputStream(String[] lines) throws IOException {
        return new ByteArrayInputStream(stringLinesToByteArray(lines));
    }
    
    public void testStudentEquals() {
        Student student = new StudentCatalogImpl().newStudent("Test student");
        assertNotNull("Metoda newStudent vraci null",student);

        assertTrue("Metoda equals vraci false, i kdyz je jejim parametrem " +
        "instance studenta se stejnym id (chyba v metode equals)",
        student.equals(new TestStudent(student.getId())));
        assertFalse("Metoda equals vraci true, i kdyz je jejim parametrem " +
        "instance studenta s odlisnym id (chyba v metode equals)",
        student.equals(new TestStudent(student.getId() + 1)));
        
        assertFalse("Metoda StudentImpl.hashCode() neni prekryta, takze porusuje " +
        "obecny kontrakt metody Object.hashCode()",student.hashCode() ==
        System.identityHashCode(student));
        
        StudentCatalog catalog = new StudentCatalogImpl();
        Student student2 = catalog.newStudent("Test student");
        assertEquals("Metoda StudentCatalogImpl.newStudent se chova " +
        "nedeterministicky; prvni vytvoreny student ma pokazde jine id",
        student.getId(),student2.getId());
        
        assertEquals("Metoda StudentImpl.hashCode() vraci pro dva stejne studenty " +
        "ruznou hodnotu, takze porusuje obecny kontrakt metody Object.hashCode() ",
        student.hashCode(),student2.hashCode());

        assertFalse("Metoda StudentImpl.hashCode() vraci pro dva ruzne studenty " +
        "stejny kod; to sice neporusuje obecny kontrakt metody Object.hashCode(), " +
        "ale v tomto pripade lze snadno vytvorit bezkolizni implementaci teto metody",
        student.hashCode() == catalog.newStudent("Test student 2").hashCode());        
    }

    public void testNewStudent() {
        StudentCatalog catalog = new StudentCatalogImpl();
        Student student = catalog.newStudent("Test student");
        assertNotNull("Metoda newStudent vraci null",student);

        assertEquals("Nove vytvoreny student ma jine jmeno, nez bylo zadano " +
        "jako parametr metody StudentCatalog.newStudent()","Test student",
        student.getName());
        
        Student student2 = catalog.getStudent(student.getId());
        assertNotNull("Metoda getStudent s identifikatorem nove vytvoreneho " +
        "studenta vraci null",student2);
        
        assertTrue("Mnozina studentu vracena metodou getStudents() neobsahuje " +
        "nove vytvoreneho studenta",catalog.getStudents().contains(student));
    }
    
    private static final String[] TEST_DATA = { 
        "1 Petr Adamek",
        "2 Mario Martinez Chuares Ferdinando",
        "3 Don Quijote De La Mancha",
        "12384 Johanka z Arku",
        "4 Frantisek Zlutoucky-Konicek"
    };
    
    public void testLoad() throws IOException, StudentCatalogException {

        StudentCatalog catalog = new StudentCatalogImpl();
        assertTrue("Katalog po vytvoreni neni prazdny",catalog.getStudents().isEmpty());
        catalog.load(createTestInputStream(TEST_DATA));
        
        assertEquals("Katalog po nacteni dat obsahuje nespravny pocet studentu", 
        TEST_DATA.length, catalog.getStudents().size());
        
        Map<Integer,String> expectedStudents = new TreeMap<Integer,String>();
        Map<Integer,String> loadedStudents = new TreeMap<Integer,String>();
 
        for (String line : TEST_DATA) {
            String[] columns = line.split(" ",2);
            int id = Integer.parseInt(columns[0]);
            String name = columns[1];
            expectedStudents.put(id,name);
            
            Student student = catalog.getStudent(id);
            assertNotNull("Metoda getStudent nenalezne studenta s id " + id + 
            ", ktery je obsazen v testovacim souboru",student);
            assertEquals("Student s id " + id + "vraceny metodou getStudent " +
            "obsahuje jine jmeno, nez jake ma v testovacim souboru",
            name,student.getName());
        }
        
        for (Student student : catalog.getStudents()) {
            loadedStudents.put(student.getId(),student.getName());
        }

        assertEquals("Mnozina nactenych studentu se lisi od ocekavanych dat",
        expectedStudents, loadedStudents);
        
        int nextStudentId = catalog.newStudent("test x").getId();
        assertFalse("Metoda newStudent pridelila novemu studentovi id (" + 
        nextStudentId +"), ktere jiz ma nektery ze studentu v testovacim souboru" +
        "pridelene",expectedStudents.keySet().contains(nextStudentId));
        
    }

    public void testSave() throws IOException, StudentCatalogException {

        StudentCatalog catalog = new StudentCatalogImpl();
        catalog.load(createTestInputStream(TEST_DATA));
        
        ByteArrayOutputStream outputStream = new ByteArrayOutputStream();
        catalog.save(outputStream);
        
        
        // Testovaci data ulozit do pomocne kolekce
        Set<String> testDataLines = new HashSet<String>();
        for (String line : TEST_DATA) {
            testDataLines.add(line);
        }
        
        BufferedReader reader = new BufferedReader(new InputStreamReader(
        new ByteArrayInputStream(outputStream.toByteArray()), ENCODING));

        String line;
        while (null != (line = reader.readLine())) {
            assertTrue("Vystupni soubor obsahuje radek, ktery nebyl ve vstupnich" +
            " testovacich datech: " + line,testDataLines.contains(line));
            testDataLines.remove(line);
        }
        assertTrue("Ve vystupnim souboru chybi nektere radky ze vstupnich " +
        "testovacich dat: " + testDataLines,testDataLines.isEmpty());
        
    }

    public void testBadDataEmptyLine() throws IOException, StudentCatalogException {
        
        StudentCatalog catalog = new StudentCatalogImpl();
        // prazdny radek
        try {
            catalog.load(createTestInputStream(new String[] { "" }));        
            fail("Metoda load nevyvola vyjimku StudentCatalogException, " +
            "ackoliv je ve vstupnim souboru prazdny radek");
        }
        catch(StudentCatalogException ex) {
            // Vsechno je v poradku
            // Toto je jediny pripad, kdy muze byt blok catch prazdny!!!
            // kdekoliv jinde je to hruba chyba!
        }
        catch(Exception ex) {
            // jakakoliv jina vyjimka je chybou
            fail("Metoda load na prazdnem radku vyvola misto vyjimky " +
            "StudentCatalogException vyjimku " + ex.getClass().getName());
        }        
    }

    public void testBadDataNoName() throws IOException, StudentCatalogException {
        
        StudentCatalog catalog = new StudentCatalogImpl();
        // chybi jmeno studenta 
        try {
            catalog.load(createTestInputStream(new String[] { "128" }));        
            fail("Metoda load nevyvola vyjimku StudentCatalogException, " +
            "ackoliv je ve vstupnim souboru radek bez jmena studenta");
        }
        catch(StudentCatalogException ex) {
            // Vsechno je v poradku
            // Toto je jediny pripad, kdy muze byt blok catch prazdny!!!
            // kdekoliv jinde je to hruba chyba!
        }
        catch(Exception ex) {
            // jakakoliv jina vyjimka je chybou
            fail("Metoda load na radku bez jmena studenta vyvola misto vyjimky " +
            "StudentCatalogException vyjimku " + ex.getClass().getName());
        }        
    }

    public void testBadDataInvalidNumber() throws IOException, StudentCatalogException {
        
        StudentCatalog catalog = new StudentCatalogImpl();
        // chybi jmeno studenta 
        try {
            catalog.load(createTestInputStream(new String[] { "1a jmeno" }));        
            fail("Metoda load nevyvola vyjimku StudentCatalogException, " +
            "ackoliv je ve vstupnim souboru radek neciselnym id");
        }
        catch(StudentCatalogException ex) {
            // Vsechno je v poradku
            // Toto je jediny pripad, kdy muze byt blok catch prazdny!!!
            // kdekoliv jinde je to hruba chyba!
        }
        catch(Exception ex) {
            // jakakoliv jina vyjimka je chybou
            fail("Metoda load na radku s neciselnym id vyvola misto vyjimky " +
            "StudentCatalogException vyjimku " + ex.getClass().getName());
        }        
    }
    
    public void testReadOnlyStudentsCollection() throws IOException, StudentCatalogException {
        
        StudentCatalog catalog = new StudentCatalogImpl();
        Collection<Student> students = catalog.getStudents();
        // chybi jmeno studenta 
        try {
            students.clear();
            fail("Metoda getStudents() vraci kolekci, kterou je mozne modifikovat");
        }
        catch(UnsupportedOperationException ex) {
            // Vsechno je v poradku
            // Toto je jediny pripad, kdy muze byt blok catch prazdny!!!
            // kdekoliv jinde je to hruba chyba!
        }
    }
    
    public void testClear() {
        StudentCatalog catalog = new StudentCatalogImpl();
        catalog.newStudent("test");
        assertFalse("Po pridani studenta je katalog stale prazdny",catalog.getStudents().isEmpty());
        catalog.clear();
        assertTrue("Po volani metody clear neni katalog prazdny",catalog.getStudents().isEmpty());
    }    
    
    public void testLoadIOError() {
        try {
            new StudentCatalogImpl().load(new BadInputStream());
            fail("Vyjimka IOException pri cteni ze vstupniho proudu" +
            " je zrejme ignorovana!!! " +
            "(nezpusobi vyvolani pozadovane vyjimky StudentCatalogException)");
        }
        catch (StudentCatalogException ex) {
            assertEquals("Vyjimka IOException pri cteni ze vstupniho proudu " +
            "sice zpusobi vyvolani vyjimky StudentCatalogException, ale tato " +
            "vyjimka nema spravne nastavenou svoji pricinu (viz retezeni vyjimek)",
            IOException.class,ex.getCause().getClass());
        }
        catch (Exception ex) {
            fail("Vyjimka IOException pri cteni ze vstupniho proudu" +
            " zpusobi vyvolani vyjimky " + ex.getClass() + " misto pozadovane " +
            "StudentCatalogException");
        }
    }

    public void testSaveIOError() {
        StudentCatalog catalog = new StudentCatalogImpl();
        catalog.newStudent("Test student");
        try {
            catalog.save(new BadOutputStream());
            fail("Vyjimka IOException pri zapisu do vystupniho proudu" +
            " je zrejme ignorovana!!! " +
            "(nezpusobi vyvolani pozadovane vyjimky StudentCatalogException). " +
            "Nepouzivate nahodou tridu PrintWriter, ktera IOException " +
            "negeneruje?");
        }
        catch (StudentCatalogException ex) {
            assertEquals("Vyjimka IOException pri zapisu do vystupniho proudu " +
            "sice zpusobi vyvolani vyjimky StudentCatalogException, ale tato " +
            "vyjimka nema spravne nastavenou svoji pricinu (viz retezeni vyjimek)",
            IOException.class,ex.getCause().getClass());
        }
        catch (Exception ex) {
            fail("Vyjimka IOException pri zapisu do vystupniho proudu" +
            " zpusobi vyvolani vyjimky " + ex.getClass() + " misto pozadovane " +
            "StudentCatalogException");
        }
    }
    
    private static class TestStudent implements Student {

        private int id;
        
        private TestStudent(int id) {
            this.id = id;
        }
        
        public String getName() {
            return "Student cislo " + id;
        }

        public int getId() {
            return id;
        }
        
    }
    
    private class BadInputStream extends InputStream {
        public int read() throws IOException {
            throw new IOException("It is not possible to read from this stream");
        }
    }
    
    private class BadOutputStream extends OutputStream {
        public void write(int value) throws IOException {
            throw new IOException("It is not possible to write to this stream");
        }
    }
    
    public static void main(String[] args) {
        junit.textui.TestRunner.run(StudentCatalogTest.class);
    }

}
